@@ -5,7 +5,9 @@ import time
 from kazoo.exceptions import (
     ConnectionClosedError,
     ConnectionLoss,
-    SessionExpiredError
+    KazooException,
+    SessionExpiredError,
+    OperationTimeoutError
 )
 
 log = logging.getLogger(__name__)
@@ -15,19 +17,27 @@ class ForceRetryError(Exception):
     """Raised when some recipe logic wants to force a retry"""
 
 
+class RetryFailedError(KazooException):
+    """Raised when retrying an operation ultimately failed, after
+    retrying the maximum number of attempts.
+    """
+
+
 class RetrySleeper(object):
     """A retry sleeper that will track its jitter, backoff and
-    sleep appropriately when asked"""
+    sleep appropriately when asked."""
     def __init__(self, max_tries=1, delay=0.1, backoff=2, max_jitter=0.8,
-                 ignore_expire=True, sleep_func=time.sleep):
+                 max_delay=3600, sleep_func=time.sleep):
         """Create a :class:`KazooRetry` instance
 
         :param max_tries: How many times to retry the command.
-        :param delay: Initial delay between retry attempts
+        :param delay: Initial delay between retry attempts.
         :param backoff: Backoff multiplier between retry attempts.
                         Defaults to 2 for exponential backoff.
         :param max_jitter: Additional max jitter period to wait between
                            retry attempts to avoid slamming the server.
+        :param max_delay: Maximum delay in seconds, regardless of other
+                          backoff settings. Defaults to one hour.
 
         """
         self.sleep_func = sleep_func
@@ -35,6 +45,7 @@ class RetrySleeper(object):
         self.delay = delay
         self.backoff = backoff
         self.max_jitter = int(max_jitter * 100)
+        self.max_delay = float(max_delay)
         self._attempts = 0
         self._cur_delay = delay
 
@@ -47,16 +58,17 @@ class RetrySleeper(object):
         """Increment the failed count, and sleep appropriately before
         continuing"""
         if self._attempts == self.max_tries:
-            raise Exception("Too many retry attempts")
+            raise RetryFailedError("Too many retry attempts")
         self._attempts += 1
         jitter = random.randint(0, self.max_jitter) / 100.0
         self.sleep_func(self._cur_delay + jitter)
-        self._cur_delay *= self.backoff
+        self._cur_delay = min(self._cur_delay * self.backoff, self.max_delay)
 
     def copy(self):
         """Return a clone of this retry sleeper"""
         return RetrySleeper(self.max_tries, self.delay, self.backoff,
-                           self.max_jitter / 100.0, self.sleep_func)
+                           self.max_jitter / 100.0, self.max_delay,
+                           self.sleep_func)
 
 
 class KazooRetry(object):
@@ -64,6 +76,7 @@ class KazooRetry(object):
     exceptions"""
     RETRY_EXCEPTIONS = (
         ConnectionLoss,
+        OperationTimeoutError,
         ForceRetryError
     )
 
@@ -72,22 +85,24 @@ class KazooRetry(object):
     )
 
     def __init__(self, max_tries=1, delay=0.1, backoff=2, max_jitter=0.8,
-                 ignore_expire=True, sleep_func=time.sleep):
+                 max_delay=3600, ignore_expire=True, sleep_func=time.sleep):
         """Create a :class:`KazooRetry` instance
 
         :param max_tries: How many times to retry the command.
-        :param delay: Initial delay between retry attempts
+        :param delay: Initial delay between retry attempts.
         :param backoff: Backoff multiplier between retry attempts.
                         Defaults to 2 for exponential backoff.
         :param max_jitter: Additional max jitter period to wait between
                            retry attempts to avoid slamming the server.
+        :param max_delay: Maximum delay in seconds, regardless of other
+                          backoff settings. Defaults to one hour.
         :param ignore_expire:
             Whether a session expiration should be ignored and treated
             as a retry-able command.
 
         """
         self.retry_sleeper = RetrySleeper(max_tries, delay, backoff, max_jitter,
-                                         sleep_func)
+                                         max_delay, sleep_func)
         self.sleep_func = sleep_func
         self.retry_exceptions = self.RETRY_EXCEPTIONS
         if ignore_expire:
