@@ -1,7 +1,12 @@
+import atexit
+import os
+import sys
 import threading
+import time
 import uuid
 import unittest
 
+from nose import SkipTest
 from nose.tools import eq_
 from nose.tools import raises
 
@@ -13,6 +18,55 @@ from kazoo.exceptions import NoNodeError
 from kazoo.exceptions import NoAuthError
 from kazoo.exceptions import ConnectionLoss
 
+if sys.version_info > (3, ):  # pragma: nocover
+    def u(s):
+        return s
+else:  # pragma: nocover
+    def u(s):
+        return unicode(s, "unicode_escape")
+
+
+class TestSessions(unittest.TestCase):
+
+    def setUp(self):
+        from kazoo.client import KazooClient
+        from kazoo.protocol.states import KazooState
+        from kazoo.testing.common import ZookeeperCluster
+        ZK_HOME = os.environ.get("ZOOKEEPER_PATH")
+        self.cluster = ZookeeperCluster(ZK_HOME, size=1, port_offset=21000)
+        self.cluster.start()
+        atexit.register(lambda cluster: self.cluster.terminate(), self.cluster)
+        self.client = KazooClient(self.cluster[0].address, max_retries=3)
+        self.ev = threading.Event()
+
+        def back(state):
+            if state == KazooState.CONNECTED:
+                self.ev.set()
+        self.client.start()
+        self.path = self.client.create(uuid.uuid4().hex)
+        self.client.add_listener(back)
+
+    def test_restarted_server(self):
+        self.cluster.stop()
+        self.cluster.start()
+        self.ev.wait(5)
+        raise SkipTest('Patch missing')
+        eq_(self.ev.is_set(), True)
+        self.assertTrue(self.client.retry(self.client.exists, self.path))
+
+    def test_terminated_server(self):
+        self.cluster.reset()
+        self.cluster.start()
+        self.ev.wait(5)
+        raise SkipTest('Patch missing')
+        eq_(self.ev.is_set(), True)
+        self.assertFalse(self.client.retry(self.client.exists, self.path))
+
+    def tearDown(self):
+        self.ev.clear()
+        self.client.stop()
+        self.cluster.stop()
+
 
 class TestClientTransitions(KazooTestCase):
     def test_connection_and_disconnection(self):
@@ -66,6 +120,17 @@ class TestClientConstructor(unittest.TestCase):
         self.assertTrue(client.handler.timeout_exception is TimeoutError)
         self.assertRaises(TimeoutError, client.start, 0.1)
 
+    def test_ordered_host_selection(self):
+        client = self._makeOne(hosts='127.0.0.1:9,127.0.0.2:9/a',
+            randomize_hosts=False)
+        hosts = [h for h in client.hosts]
+        eq_(hosts, [('127.0.0.1', 9), ('127.0.0.2', 9)])
+
+    def test_invalid_hostname(self):
+        client = self._makeOne(hosts='nosuchhost/a')
+        timeout = client.handler.timeout_exception
+        self.assertRaises(timeout, client.start, 0.1)
+
 
 class TestConnection(KazooTestCase):
     def _makeAuth(self, *args, **kwargs):
@@ -80,17 +145,16 @@ class TestConnection(KazooTestCase):
         acl = self._makeAuth(username, password, all=True)
 
         self.client.add_auth("digest", digest_auth)
-
         self.client.default_acl = (acl,)
 
-        self.client.create("/1")
-        self.client.create("/1/2")
-        self.client.ensure_path("/1/2/3")
-
-        eve = self._get_client()
-        eve.start()
-
         try:
+            self.client.create("/1")
+            self.client.create("/1/2")
+            self.client.ensure_path("/1/2/3")
+
+            eve = self._get_client()
+            eve.start()
+
             self.assertRaises(NoAuthError, eve.get, "/1/2")
 
             # try again with the wrong auth token
@@ -111,12 +175,40 @@ class TestConnection(KazooTestCase):
         client = self._get_client(auth_data=[('digest', digest_auth)])
         client.start()
         try:
-            client.create('/1', "", acl=(acl,))
+            client.create('/1', acl=(acl,))
+            # give ZK a chance to copy data to other node
+            time.sleep(0.1)
             self.assertRaises(NoAuthError, self.client.get, "/1")
         finally:
             client.delete('/1')
             client.stop()
 
+    def test_unicode_auth(self):
+        username = u("xe4/\hm")
+        password = u("/\xe4hm")
+        digest_auth = "%s:%s" % (username, password)
+        acl = self._makeAuth(username, password, all=True)
+
+        self.client.add_auth("digest", digest_auth)
+        self.client.default_acl = (acl,)
+
+        try:
+            self.client.create("/1")
+            self.client.ensure_path("/1/2/3")
+
+            eve = self._get_client()
+            eve.start()
+
+            self.assertRaises(NoAuthError, eve.get, "/1/2")
+
+            # try again with the wrong auth token
+            eve.add_auth("digest", "badbad:bad")
+
+            self.assertRaises(NoAuthError, eve.get, "/1/2")
+        finally:
+            # Ensure we remove the ACL protected nodes
+            self.client.delete("/1", recursive=True)
+
     def test_invalid_auth(self):
         self.assertRaises(TypeError, self.client.add_auth,
             'digest', ('user', 'pass'))
@@ -244,10 +336,10 @@ class TestClient(KazooTestCase):
 
     def test_create_unicode_path(self):
         client = self.client
-        path = client.create(u"/ascii")
-        eq_(path, u"/ascii")
-        path = client.create(u"/\xe4hm")
-        eq_(path, u"/\xe4hm")
+        path = client.create(u("/ascii"))
+        eq_(path, u("/ascii"))
+        path = client.create(u("/\xe4hm"))
+        eq_(path, u("/\xe4hm"))
 
     def test_create_invalid_path(self):
         client = self.client
@@ -269,20 +361,20 @@ class TestClient(KazooTestCase):
 
     def test_create_value(self):
         client = self.client
-        client.create("/1", "bytes")
+        client.create("/1", b"bytes")
         data, stat = client.get("/1")
-        eq_(data, "bytes")
+        eq_(data, b"bytes")
 
     def test_create_unicode_value(self):
         client = self.client
-        self.assertRaises(TypeError, client.create, "/1", u"\xe4hm")
+        self.assertRaises(TypeError, client.create, "/1", u("\xe4hm"))
 
     def test_create_large_value(self):
         client = self.client
-        kb_512 = "a" * (512 * 1024)
+        kb_512 = b"a" * (512 * 1024)
         client.create("/1", kb_512)
         self.assertTrue(client.exists("/1"))
-        mb_2 = "a" * (2 * 1024 * 1024)
+        mb_2 = b"a" * (2 * 1024 * 1024)
         self.assertRaises(ConnectionLoss, client.create, "/2", mb_2)
 
     def test_create_acl_duplicate(self):
@@ -304,71 +396,71 @@ class TestClient(KazooTestCase):
 
     def test_create_ephemeral(self):
         client = self.client
-        client.create("/1", "ephemeral", ephemeral=True)
+        client.create("/1", b"ephemeral", ephemeral=True)
         data, stat = client.get("/1")
-        eq_(data, "ephemeral")
+        eq_(data, b"ephemeral")
         eq_(stat.ephemeralOwner, client.client_id[0])
 
     def test_create_no_ephemeral(self):
         client = self.client
-        client.create("/1", "val1")
+        client.create("/1", b"val1")
         data, stat = client.get("/1")
         self.assertFalse(stat.ephemeralOwner)
 
     def test_create_ephemeral_no_children(self):
         from kazoo.exceptions import NoChildrenForEphemeralsError
         client = self.client
-        client.create("/1", "ephemeral", ephemeral=True)
+        client.create("/1", b"ephemeral", ephemeral=True)
         self.assertRaises(NoChildrenForEphemeralsError,
-            client.create, "/1/2", "val1")
+            client.create, "/1/2", b"val1")
         self.assertRaises(NoChildrenForEphemeralsError,
-            client.create, "/1/2", "val1", ephemeral=True)
+            client.create, "/1/2", b"val1", ephemeral=True)
 
     def test_create_sequence(self):
         client = self.client
-        client.create("/folder", "")
-        path = client.create("/folder/a", "sequence", sequence=True)
+        client.create("/folder")
+        path = client.create("/folder/a", b"sequence", sequence=True)
         eq_(path, "/folder/a0000000000")
-        path2 = client.create("/folder/a", "sequence", sequence=True)
+        path2 = client.create("/folder/a", b"sequence", sequence=True)
         eq_(path2, "/folder/a0000000001")
 
     def test_create_ephemeral_sequence(self):
         basepath = "/" + uuid.uuid4().hex
-        realpath = self.client.create(basepath, "sandwich", sequence=True,
+        realpath = self.client.create(basepath, b"sandwich", sequence=True,
             ephemeral=True)
         self.assertTrue(basepath != realpath and realpath.startswith(basepath))
         data, stat = self.client.get(realpath)
-        eq_(data, "sandwich")
+        eq_(data, b"sandwich")
 
     def test_create_makepath(self):
-        self.client.create("/1/2", "val1", makepath=True)
+        self.client.create("/1/2", b"val1", makepath=True)
         data, stat = self.client.get("/1/2")
-        eq_(data, "val1")
+        eq_(data, b"val1")
 
-        self.client.create("/1/2/3/4/5", "val2", makepath=True)
+        self.client.create("/1/2/3/4/5", b"val2", makepath=True)
         data, stat = self.client.get("/1/2/3/4/5")
-        eq_(data, "val2")
+        eq_(data, b"val2")
 
     def test_create_no_makepath(self):
-        self.assertRaises(NoNodeError, self.client.create, "/1/2", "val1")
-        self.assertRaises(NoNodeError, self.client.create, "/1/2", "val1",
+        self.assertRaises(NoNodeError, self.client.create, "/1/2", b"val1")
+        self.assertRaises(NoNodeError, self.client.create, "/1/2", b"val1",
             makepath=False)
 
     def test_create_exists(self):
         from kazoo.exceptions import NodeExistsError
         client = self.client
-        path = client.create("/1", "")
-        self.assertRaises(NodeExistsError, client.create, path, "")
+        path = client.create("/1")
+        self.assertRaises(NodeExistsError, client.create, path)
 
     def test_create_get_set(self):
         nodepath = "/" + uuid.uuid4().hex
 
-        self.client.create(nodepath, "sandwich", ephemeral=True)
+        self.client.create(nodepath, b"sandwich", ephemeral=True)
 
         data, stat = self.client.get(nodepath)
-        eq_(data, "sandwich")
+        eq_(data, b"sandwich")
 
-        newstat = self.client.set(nodepath, "hats", stat.version)
+        newstat = self.client.set(nodepath, b"hats", stat.version)
         self.assertTrue(newstat)
         assert newstat.version > stat.version
 
@@ -411,7 +503,7 @@ class TestClient(KazooTestCase):
         exists = self.client.exists(nodepath)
         eq_(exists, None)
 
-        self.client.create(nodepath, "sandwich", ephemeral=True)
+        self.client.create(nodepath, b"sandwich", ephemeral=True)
         exists = self.client.exists(nodepath)
         self.assertTrue(exists)
         assert isinstance(exists.version, int)
@@ -427,8 +519,7 @@ class TestClient(KazooTestCase):
 
     def test_exists_watch(self):
         nodepath = "/" + uuid.uuid4().hex
-
-        event = threading.Event()
+        event = self.client.handler.event_object()
 
         def w(watch_event):
             eq_(watch_event.path, nodepath)
@@ -437,15 +528,14 @@ class TestClient(KazooTestCase):
         exists = self.client.exists(nodepath, watch=w)
         eq_(exists, None)
 
-        self.client.create(nodepath, "x", ephemeral=True)
+        self.client.create(nodepath, ephemeral=True)
 
         event.wait(1)
         self.assertTrue(event.is_set())
 
     def test_exists_watcher_exception(self):
         nodepath = "/" + uuid.uuid4().hex
-
-        event = threading.Event()
+        event = self.client.handler.event_object()
 
         # if the watcher throws an exception, all we can really do is log it
         def w(watch_event):
@@ -457,7 +547,7 @@ class TestClient(KazooTestCase):
         exists = self.client.exists(nodepath, watch=w)
         eq_(exists, None)
 
-        self.client.create(nodepath, "x", ephemeral=True)
+        self.client.create(nodepath, ephemeral=True)
 
         event.wait(1)
         self.assertTrue(event.is_set())
@@ -465,7 +555,7 @@ class TestClient(KazooTestCase):
     def test_create_delete(self):
         nodepath = "/" + uuid.uuid4().hex
 
-        self.client.create(nodepath, "zzz")
+        self.client.create(nodepath, b"zzz")
 
         self.client.delete(nodepath)
 
@@ -518,18 +608,18 @@ class TestClient(KazooTestCase):
 
     def test_set(self):
         client = self.client
-        client.create('a', 'first')
-        stat = client.set('a', 'second')
+        client.create('a', b'first')
+        stat = client.set('a', b'second')
         data, stat2 = client.get('a')
-        self.assertEqual(data, 'second')
+        self.assertEqual(data, b'second')
         self.assertEqual(stat, stat2)
 
     def test_set_invalid_arguments(self):
         client = self.client
-        client.create('a', 'first')
-        self.assertRaises(TypeError, client.set, ('a', 'b'), 'value')
+        client.create('a', b'first')
+        self.assertRaises(TypeError, client.set, ('a', 'b'), b'value')
         self.assertRaises(TypeError, client.set, 'a', ['v', 'w'])
-        self.assertRaises(TypeError, client.set, 'a', 'value', 'V1')
+        self.assertRaises(TypeError, client.set, 'a', b'value', 'V1')
 
     def test_delete(self):
         client = self.client
@@ -617,6 +707,11 @@ class TestClient(KazooTestCase):
             client._safe_close()
         testit()
 
+    def test_client_state(self):
+        from kazoo.protocol.states import KeeperState
+        eq_(self.client.client_state, KeeperState.CONNECTED)
+
+
 dummy_dict = {
     'aversion': 1, 'ctime': 0, 'cversion': 1,
     'czxid': 110, 'dataLength': 1, 'ephemeralOwner': 'ben',
@@ -624,6 +719,142 @@ dummy_dict = {
 }
 
 
+class TestClientTransactions(KazooTestCase):
+    def setUp(self):
+        KazooTestCase.setUp(self)
+        ver = self.client.server_version()
+        if ver[1] < 4:
+            raise SkipTest("Must use zookeeper 3.4 or above")
+
+    def test_basic_create(self):
+        t = self.client.transaction()
+        t.create('/freddy')
+        t.create('/fred', ephemeral=True)
+        t.create('/smith', sequence=True)
+        results = t.commit()
+        eq_(results[0], '/freddy')
+        eq_(len(results), 3)
+        self.assertTrue(results[2].startswith('/smith0'))
+
+    def test_bad_creates(self):
+        args_list = [(True,), ('/smith', 0), ('/smith', b'', 'bleh'),
+                     ('/smith', b'', None, 'fred'),
+                     ('/smith', b'', None, True, 'fred')]
+
+        @raises(TypeError)
+        def testit(args):
+            t = self.client.transaction()
+            t.create(*args)
+
+        for args in args_list:
+            testit(args)
+
+    def test_default_acl(self):
+        from kazoo.security import make_digest_acl
+        username = uuid.uuid4().hex
+        password = uuid.uuid4().hex
+
+        digest_auth = "%s:%s" % (username, password)
+        acl = make_digest_acl(username, password, all=True)
+
+        self.client.add_auth("digest", digest_auth)
+        self.client.default_acl = (acl,)
+
+        t = self.client.transaction()
+        t.create('/freddy')
+        results = t.commit()
+        eq_(results[0], '/freddy')
+
+    def test_basic_delete(self):
+        self.client.create('/fred')
+        t = self.client.transaction()
+        t.delete('/fred')
+        results = t.commit()
+        eq_(results[0], True)
+
+    def test_bad_deletes(self):
+        args_list = [(True,), ('/smith', 'woops'), ]
+
+        @raises(TypeError)
+        def testit(args):
+            t = self.client.transaction()
+            t.delete(*args)
+
+        for args in args_list:
+            testit(args)
+
+    def test_set(self):
+        self.client.create('/fred', b'01')
+        t = self.client.transaction()
+        t.set_data('/fred', b'oops')
+        t.commit()
+        res = self.client.get('/fred')
+        eq_(res[0], b'oops')
+
+    def test_bad_sets(self):
+        args_list = [(42, 52), ('/smith', False), ('/smith', b'', 'oops')]
+
+        @raises(TypeError)
+        def testit(args):
+            t = self.client.transaction()
+            t.set_data(*args)
+
+        for args in args_list:
+            testit(args)
+
+    def test_check(self):
+        self.client.create('/fred')
+        version = self.client.get('/fred')[1].version
+        t = self.client.transaction()
+        t.check('/fred', version)
+        t.create('/blah')
+        results = t.commit()
+        eq_(results[0], True)
+        eq_(results[1], '/blah')
+
+    def test_bad_checks(self):
+        args_list = [(42, 52), ('/smith', 'oops')]
+
+        @raises(TypeError)
+        def testit(args):
+            t = self.client.transaction()
+            t.check(*args)
+
+        for args in args_list:
+            testit(args)
+
+    def test_bad_transaction(self):
+        from kazoo.exceptions import RolledBackError, NoNodeError
+        t = self.client.transaction()
+        t.create('/fred')
+        t.delete('/smith')
+        results = t.commit()
+        eq_(results[0].__class__, RolledBackError)
+        eq_(results[1].__class__, NoNodeError)
+
+    def test_bad_commit(self):
+        t = self.client.transaction()
+
+        @raises(ValueError)
+        def testit():
+            t.commit()
+
+        t.committed = True
+        testit()
+
+    def test_bad_context(self):
+        @raises(TypeError)
+        def testit():
+            with self.client.transaction() as t:
+                t.check(4232)
+        testit()
+
+    def test_context(self):
+        with self.client.transaction() as t:
+            t.create('/smith', b'32')
+        eq_(self.client.get('/smith')[0], b'32')
+
+
 class TestCallbacks(unittest.TestCase):
     def test_session_callback_states(self):
         from kazoo.protocol.states import KazooState, KeeperState
