@@ -26,12 +26,23 @@ class TestRetrySleeper(unittest.TestCase):
             retry.increment()
         testit()
 
+    def test_maximum_delay(self):
+        def sleep_func(time):
+            pass
+
+        retry = self._makeOne(delay=10, max_tries=100, sleep_func=sleep_func)
+        for i in range(10):
+            retry.increment()
+        self.assertTrue(retry._cur_delay < 4000, retry._cur_delay)
+        # gevent's sleep function is picky about the type
+        eq_(type(retry._cur_delay), float)
+
 
 class TestKazooRetry(unittest.TestCase):
 
-    def _makeOne(self):
+    def _makeOne(self, **kw):
         from kazoo.retry import KazooRetry
-        return KazooRetry()
+        return KazooRetry(**kw)
 
     def test_connection_closed(self):
         from kazoo.exceptions import ConnectionClosedError
@@ -40,3 +51,11 @@ class TestKazooRetry(unittest.TestCase):
         def testit():
             raise ConnectionClosedError()
         self.assertRaises(ConnectionClosedError, retry, testit)
+
+    def test_session_expired(self):
+        from kazoo.exceptions import SessionExpiredError
+        retry = self._makeOne(max_tries=1)
+
+        def testit():
+            raise SessionExpiredError()
+        self.assertRaises(Exception, retry, testit)
